//! Generate register interface rust code from types in [`crate::types`].

use crate::{
    type_traits::GetName,
    types::{Block, Module, ModuleBlockElements},
};
use heck::{ToSnakeCase, ToUpperCamelCase};
use proc_macro2::TokenStream;
use quote::quote;

pub trait CodeGen {
    fn generate_register_interface(self) -> proc_macro2::TokenStream;
}

impl CodeGen for Module {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        let mut out = TokenStream::new();
        if !self.elements_bitstring.is_empty() {
            todo!("bitstring generation is not yet implemented")
        }

        let child_mods = self
            .elements_other
            .into_iter()
            .map(|e| e.generate_register_interface());
        for child in child_mods {
            out.extend(child);
        }

        out
    }
}

impl CodeGen for ModuleBlockElements {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        match self {
            ModuleBlockElements::Block(block) => block.generate_register_interface(),
            ModuleBlockElements::Register(register) => todo!(),
            ModuleBlockElements::Memory(memory) => todo!(),
            ModuleBlockElements::Fifo(fifo) => todo!(),
        }
    }
}

impl CodeGen for Block {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        let snake_case_name = self.name.to_snake_case();
        let upper_camel_name = self.name.to_upper_camel_case();
        let addr = format!("0x{:x}", self.addr);
        let desc = self.desc.unwrap_or("".to_string());

        let accessor_methods = {
            let mut out = TokenStream::new();
            for child_name in self.elements.iter().map(|e| e.get_name()) {
                let snake_case_name = child_name.to_snake_case();
                let upper_camel_name = child_name.to_upper_camel_case();

                out.extend(quote! {
                    pub fn #snake_case_name(&self) -> #snake_case_name::#upper_camel_name {
                        #snake_case_name::#upper_camel_name::new(self.mem_ptr)
                    }
                });
            }

            out
        };
 
        let child_mods = {
            let mut out = TokenStream::new();
            for child_mod in self.elements.into_iter() {
                let module = child_mod.generate_register_interface();
                out.extend(module);
            }

            out
        };

        quote! {
            pub mod #snake_case_name {
                #[doc = #desc]

                use std::marker::PhantomData;

                use super::RegisterInterface;

                #child_mods

                const OFFSET: usize = #addr;
                //pub(crate) const SIZE: usize = 0x4000;

                /// Access to SLR0.
                pub struct #upper_camel_name<'a> {
                    mem_ptr: *mut u32,
                    _marker: PhantomData<&'a mut RegisterInterface>,
                }

                impl #upper_camel_name<'_> {
                    pub(crate) fn new(parent_ptr: *mut u32) -> Self {
                        #upper_camel_name {
                            mem_ptr: unsafe { parent_ptr.add(OFFSET) },
                            _marker: PhantomData,
                        }
                    }

                    #accessor_methods
                }
            }
        }
    }
}
